#ifndef _VLP_EDITOR_H
#define _VLP_EDITOR_H

#include <QtGui/QTextEdit>
#include <QtGui/QMainWindow>
#include <QtGui/QLineEdit>
#include <QtGui/QPushButton>
#include <QtGui/QLabel>

#include "ui/editor.h"

/**
 * @file
 */

namespace loglan {
namespace vlp {

#define COMP_MODE	1
#define GEN_MODE	2
#define ALL_MODE	3

typedef struct {
	char name[255];
	char filename[255];
} FuncEntry;

typedef struct {
	QList<FuncEntry> dictionary;
	QStringList names;
	char name[255];
} CategoryEntry;

class Editor : public QMainWindow, private Ui::EditorWindow {
	Q_OBJECT
public:
	QString compiler_path;
	QString gen_path;
	QString file_path;

	Editor(int argc, char **argv);
	~Editor();

	void compile(int mode);

public:
	void load(const char *fileName);
	void save(const char *fileName);
private:
	QString fname;
	QString find_text;
	bool sensitive;
	char HomeDir[255];

private slots:
	void on_actionNew_triggered();
	void on_actionOpen_triggered();
	void on_actionSave_triggered();
	void on_actionSave_as_triggered();
	void on_actionQuit_triggered();
	void on_actionCopy_triggered();
	void on_actionPaste_triggered();
	void on_actionCut_triggered();
	void on_actionClear_all_triggered();
	void on_actionFind_triggered();
	void on_actionFind_Next_triggered();
	void on_actionPreferences_triggered();

	void on_actionCompile_triggered();
	void on_actionGen_triggered();
	void on_actionCompile_Gen_triggered();

	void on_actionProgram_structure_triggered();
	void on_actionUnit_structure_triggered();


	void on_editor_cursorPositionChanged();
};

}
}

#endif // _VLP_EDITOR_H
