/*     Loglan82 Compiler&Interpreter
     Copyright (C) 1993 Institute of Informatics, University of Warsaw
     Copyright (C)  1993, 1994 LITA, Pau
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
             You should have received a copy of the GNU General Public License
             along with this program; if not, write to the Free Software
             Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 contacts:  Andrzej.Salwicki@univ-pau.fr

or             Andrzej Salwicki
                LITA   Departement d'Informatique
                Universite de Pau
                Avenue de l'Universite
                64000 Pau   FRANCE
		 tel.  ++33 59923154	fax. ++33 59841696

=======================================================================
*/

#include "depend.h"
#include "genint.h"
#include "int.h"
#include "process.h"
#include "intproto.h"

/* Transfer of control routines */
#ifndef NO_PROTOTYPES
static void att2(virtaddr *, word, word);
static void back1(word, word, virtaddr *, word *);
#else
static void att2();
static void back1();
#endif

/**
 *Transfer control to the newly created object.
 */
void go(word ah, word am)
{
	protdescr *ptr;
	word pnum, plen, node, apt;
	message msg;

	ptr = prototype[M[am + PROTNUM]];
	apt = am + M[am];
	/* new process creation */
	if (ptr->kind == PROCESS) {
		/* save template address */
		thisp->template.addr = ah;
		thisp->template.mark = M[ah + 1];
		msg.control.type = CREATE;
		msg.control.par = M[am + PROTNUM];
		moveparams(thispix, am, &msg, PARIN, LOADPAR);
		/* pix  will create receiver */
		msg.control.receiver.pix = 0;
		/* mark will create receiver */
		msg.control.receiver.mark= 0;
		/* node we decided  */
		msg.control.receiver.node = getnode(am);
		/* send create request */
		sendmsg1(&msg);
#ifdef RPCDBG
		fprintf(stderr, "send new process from %d to node %d\n",
					thispix, msg.control.receiver.node);
#endif
		/* and wait for return from process */
		passivate(WAITFORNEW);
	} else if (isprocess((virtaddr*)(M+apt+SL))) {
		/* remote procedure call */
		/* save template address */
		thisp->backobj.addr = ah;
		thisp->backobj.mark = M[ah + 1];
		/* physical address also */
		thisp->M[temporary] = am;
		{
			virtaddr v;
			loadvirt(v, apt + SL);
			obj2mess(M, &v, &msg.control.receiver);

		/*
			fprintf(
				stderr, "send rpc from process %d to (%d,%d,%d)\n",
				thispix,
				msg.control.receiver.node,
				msg.control.receiver.pix,
				msg.control.receiver.mark
			);*/

		}
		msg.control.type = RPCALL;
		msg.control.par = M[am + PROTNUM];
		moveparams(thispix, am, &msg, PARIN, LOADPAR);
		/* send RPC request */
		sendmsg1(&msg);
		/* and wait for RP return */
		passivate(WAITFORRPC);
	} else {
		/* save local control */
		M[c1 + M[c1] + LSC] = ic;
		/* release DISPLAY */
		loosen();
		/* update DISPLAY */
		update(am, ah);
		/* new current */
		c1 = am;
		c2 = c1 + ptr->span;
		pnum = ptr->preflist;
		plen = ptr->lthpreflist;
		/* search for executable prefix */
		while (TRUE)
			if (plen <= 1) {
				ic = ptr->codeaddr;
				break;
			} else {
				ptr = prototype[M[pnum]];
				plen--;
				pnum++;
				if (ptr->kind != RECORD)
					plen = 0;
			}
	}
}


/**
 * Transfer control to a local unprefixed procedure, function, block,
 * class or coroutine.
 */
void goloc(word ah, word am)
{
	word t1;
	protdescr *ptr;

	/* save local control */
	M[c1 + M[c1] + LSC] = ic;
	/* new current */
	c1 = am;
	t1 = M[am + PROTNUM];
	ptr = prototype[t1];
	c2 = am + ptr->span;
	ic = ptr->codeaddr;
	/* simulate update display */
	M[display + t1] = am;
	M[display2 + t1] = ah;
	M[am + M[am]+STATSL ]++;
}

/** Return from block. */
void backbl(virtaddr *virt, word *am)
{
	word t1;

	t1 = M[c1 + PROTNUM];
	virt->addr = M[display2 + t1];
	/* prepare old address */
	virt->mark = M[virt->addr + 1];
	/* am of old */
	*am = c1;
	/* simulate loosen */
	M[display + t1] = 0;
	t1 = c1 + M[c1];
	/* remove from SL chain */
	M[t1 + STATSL]--;
	/* return up along SL */
	c1 = M[t1 + SL];
	/* return from main */
	if (c1 == DUMMY)
		endprocess(0);

	/* am of new current */
	c1 = M[c1];
	c2 = c1 + prototype[M[c1 + PROTNUM]]->span;
	ic = M[c1 + M[c1] + LSC];
	/* force DL consistency */
	storevirt(*virt, *am + M[*am]+DL);
}


/* Common code for some backs below. */
static void back1(word at1, word at2, virtaddr *virt, word *am)
{
	word t1;

	loosen();
	if (at1 == 0)
		endprocess(0);
	t1 = M[c1 + PROTNUM];
	/* ah of old */
	virt->addr = M[display2 + t1];
	virt->mark = M[virt->addr + 1];
	/* am of old */
	*am = c1;
	/* loop up DL */
	storevirt(*virt, at2);
	/* am of DL */
	at2 = M[at1];
	update(at2, at1);
	c1 = at2;
	c2 = c1 + prototype[M[c1 + PROTNUM]]->span;
	ic = M[c1 + M[c1] + LSC];
}


/**
 * Return from classes, coroutines and by end from procedures.
 */
void back(virtaddr *virt, word *am, word length)
	{
	word t1, t2, plist;
	int i;
	protdescr *ptr;
	message msg;

	t2 = c1 + M[c1];
	/* ah of DL */
	t1 = M[t2 + DL];
	/* prototype of current */
	ptr = prototype[M[c1 + PROTNUM]];
	/* RETURN in process */
	if (ptr->kind == PROCESS) {
#ifdef RPCDBG
		fprintf( stderr, "return from process %d\n", thispix);
#endif
		if (M[c1 + PROTNUM] == MAINBLOCK)
			endprocess(0);
		{
			virtaddr v;
			loadvirt(v, t2 + DL);
			/* father process */
			obj2mess(M, &v, &msg.control.receiver);
		}
		msg.control.type = CREACK;
		moveparams(thispix, c1, &msg, PAROUT, LOADPAR);
		/* send create acknowledge */
		sendmsg1(&msg);
		/* cut DL of new process head */
		M[t2 + DL] = 0;
		/* and suspend new process */
		passivate(STOPPED);
	} else if (ptr->kind == COROUTINE) {
	/* nothing if detached */
		if (t1 != 0) {
			M[t2 + LSC] = ic;
			back1(t1, t2 + DL, virt, am);
		}
	} else {
		/* save begining of prototype list */
		plist = ic;
		if (ptr->lthpreflist == 1 && t1 == M[t2 + SL] &&
					M[t2 + DL + 1] == M[t2 + SL + 1])
			backbl(virt, am);
		else
			back1(t1, t2 + DL, virt, am);

/*
#ifdef RPCDBG
            fprintf(
                     stderr, "back (thisp=%d) from %s to %s\n",
                     thispix,
                     (
                       (ptr->kind==PROCEDURE) ?
                         "PROCEDURE"          :
                       (ptr->kind==FUNCTION)  ?
                         "FUNCTION"           :
                         "???"
                     ),
                     isprocess((virtaddr*)(M+t2+RPCDL)) ? "PROCESS" : "OBJECT"
                   );
#endif
*/
		if ((ptr->kind == PROCEDURE || ptr->kind == FUNCTION) &&
					isprocess((virtaddr*)(M+t2+RPCDL))) {

			{
				virtaddr v;
				loadvirt(v, t2 + RPCDL);
				/* remote DL */
				obj2mess(M, &v, &msg.control.receiver); 
			}

			/*
			fprintf(stderr,
				"send rpc ack from process %d to (%d,%d,%d)\n",
				thispix,
				msg.control.receiver.node,
				msg.control.receiver.pix,
				msg.control.receiver.mark
			);
			*/
			msg.control.type = RPCACK;
			moveparams(thispix, *am, &msg, PAROUT, LOADPAR);
			/* send RP return - acknowledge */
			sendmsg1(&msg);
			/* kill procedure object manualy */
			gkill(virt);
			/* restore RPC mask from stack */
			popmask(thispix);
			/* and modify it */
			for (i = 0; i < length; i++) {
				/* prototype number */
				t1 = virtprot(M[plist++]);
				if (t1 > 0)
					enable(thispix, t1);
				else
					disable(thispix, -t1);
			}
			/* check for enabled RPCs */
			evaluaterpc(thispix);
		}
	}
}


/**
 * Return, end in procedures and functions without prefix.
 */
void backpr(virtaddr *virt, word *am)
{
	word t1, t2, t3;
	/* DL pointer of current */
	t2 = c1 + M[c1] + DL;
	/* ah of DL */
	t1 = M[t2];
	/* SL pointer */
	t3 = c1 + M[c1]+SL;
	/* SL=DL */
	if (t1 == M[t3] && M[t2 + 1] == M[t3 + 1])
		backbl(virt, am);
	else
		back1(t1, t2, virt, am);
}

/** End in classes and coroutines. */
void fin(word backic, virtaddr *virt, word *am)
{
	word t1, t2, knd;

	knd = prototype[M[c1 + PROTNUM]]->kind;
	if (knd != COROUTINE && knd != PROCESS) {
		/* a class - exit as above */
		back(virt, am, (word) 0);
	}
	else {
		/* backspace ic */
		ic = backic;
		t2 = c1 + M[c1];
		/* ah of DL */
		t1 = M[t2 + DL];
		if (t1 == 0) {
			if (M[t2 + SL] == DUMMY)
				endprocess(0);
			/* coroutine terminated */
			ic = 0;
			*am = 0;
			detach();
		} else {
			M[t2 + LSC] = ic;
			back1(t1, t2 + DL, virt, am);
		}
	}
}

/** Helper for attach/detach */
static void att2(virtaddr *virt, word ax, word at1)
{
	word t1, t2, phead;

	/* ah of current */
	t1 = M[display2 + M[c1 + PROTNUM]];
	/* DL of coroutine head */
	t2 = at1 + M[at1] + DL;
	/* loop up DL */
	M[t2] = t1;
	M[t2 + 1] = M[t1 + 1];
	/* preserve local control */
	M[c1 + M[c1] + LSC] = ic;
	loosen();
	phead = thisp->prochead;
	storevirt(*virt, phead + M[phead] + CHD);
	t2 = M[ax + DL];
	/* coroutine active */
	if (t2 == 0)
		errsignal(RTECORAC);
	/* cut DL of new coroutine head */
	M[ax + DL] = 0;
	c1 = M[t2];
	update(c1, t2);
	c2 = c1 + prototype[M[c1 + PROTNUM]]->span;
	ic = M[ c1 + M[c1] + LSC];
	/* coroutine terminated */
	if (ic == 0)
		errsignal(RTECORTM);
}


void attach(virtaddr *virt)
{
	word t1, ax, phead, chead;
	int knd;

	if (M[virt->addr + 1] != virt->mark)
		errsignal(RTEILLAT);
	else {
		/* am */
		ax = M[virt->addr];
	}
	t1 = M[ax + PROTNUM];
	if (t1 == AINT || t1 == AREAL || t1 == AVIRT || t1 == FILEOBJECT)
		errsignal(RTEILLAT);
	knd = prototype[t1]->kind;
	if (knd != COROUTINE && knd != PROCESS)
		errsignal(RTEILLAT);
	ax = ax + M[ax];
	phead = thisp->prochead;
	chead = phead + M[phead] + CHD;
	if (virt->addr != M[chead] || virt->mark != M[chead + 1]) {
		M[ax + CL] = M[chead];
		M[ax + CL + 1] = M[chead + 1];
		att2(virt, ax, M[M[chead]]);
	}
}


void detach()
{
	virtaddr virt;
	word t1, phead;

	phead = thisp->prochead;
	/* am of coroutine head */
	t1 = M[M[phead + M[phead] + CHD]];
	/* coroutine link */
	loadvirt(virt, t1 + M[t1] + CL);
	if (M[virt.addr + 1] != virt.mark)
		errsignal(RTEILLDT);
	att2(&virt, M[virt.addr] + M[M[virt.addr]], t1);
}

/* Simulate execution of inner */
void inner(word level)
{
	word t1;
	protdescr *ptr;

	/* prototype of current */
	ptr = prototype[M[c1 + PROTNUM]];
	t1 = ptr->lthpreflist;
	if (t1 != level)
		if (level == t1-1)
			ic = ptr->codeaddr;
		else
			ic = prototype[M[ptr->preflist + level]]->codeaddr;
}

