/*     Loglan82 Compiler&Interpreter
     Copyright (C) 1993 Institute of Informatics, University of Warsaw
     Copyright (C)  1993, 1994 LITA, Pau
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
             You should have received a copy of the GNU General Public License
             along with this program; if not, write to the Free Software
             Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 contacts:  Andrzej.Salwicki@univ-pau.fr

or             Andrzej Salwicki
                LITA   Departement d'Informatique
                Universite de Pau
                Avenue de l'Universite
                64000 Pau   FRANCE
		 tel.  ++33 59923154	fax. ++33 59841696

=======================================================================
*/

#include "depend.h"
#include "genint.h"
#include "int.h"
#include "process.h"
#include "intproto.h"

#include <assert.h>

/**
 * @file
 */

/** Execute one L-code instruction */
void execute()
{
	word t1, t2;
	int i;
	real r;
	virtaddr virt1, virt2, virt3;
	switch (opcode) {
	/* LOPENRC */
	case 1:
		openrc(a3, &virt2, &t2);
		storevirt(virt2, a1);
		M[a2] = t2;
		break;
	/* LBACKADDR */
	case 2:
		storevirt(thisp->backobj, a1);
		M[a2] = M[temporary];
		break;
	/* LRAISE */
	case 3:
		/* skip the address */
		ic++;
		raise_signal(a3, M[ic - 1], &t1, &t2);
		M[a1] = t1;
		M[a2] = t2;
		break;
	/* LOPEN */
	case 4:
		openobj(M[a3], &t1, &t2);
		M[a1] = t1;
		M[a2] = t2;
		break;
	/* LSLOPEN */
	case 5:
		loadvirt(virt3, a3);
		slopen(M[a3 + APREF], &virt3, &t1, &t2);
		M[a1] = t1;
		M[a2] = t2;
		break;
	/* LTHIS */
	case 15:
		virt1.addr = M[display2 + a2];
		virt1.mark = M[virt1.addr + 1];
		storevirt(virt1, a1);
		break;
	/* LVIRTDISPL */
	case 20:
		t2 = M[display + a2];
		t1 = M[t2 + PROTNUM];
		M[a1] = M[prototype[t1]->virtlist + a3];
		break;
	/* LSTATTYPE */
	case 21:
		M[a1] = a2;
		M[a1 + 1] = a3;
		break;
	/* LIPAROUT */
	case 23:
		M[a1] = param[a3].xword;
		break;
	/* LRPAROUT */
	case 24:
		MR(a1) = param[a3].xreal;
		break;
	/* LVPAROUT */
	case 25:
		storevirt(param[a3].xvirt, a1);
		break;
	/* LSIGN */
	case 31:
		if (M[a2] == 0)
			M[a1] = 0;
		else if (M[a2] < 0)
			M[a1] = -1;
		else
			M[a1] = 1;
		break;
	/* LLOWER */
	case 33:
	case 34:
		loadvirt(virt2, a2);
		if (member(&virt2, &t1)) {
			switch ((int) M[t1 + PROTNUM]) {
			case AINT:
				t2 = APINT;
				break;
			case AREAL:
				t2 = APREAL;
				break;
			case AVIRT:
				t2 = APREF;
				break;
			}
			M[a1] = (M[t1 + 2 ] + 3) / t2;
		}
		else
			errsignal(RTEREFTN);
		break;
	/* LUPPER */
	case 35:
	case 36:
		loadvirt(virt2, a2);
		if (member(&virt2, &t1)) {
			switch ((int) M[t1 + PROTNUM]) {
			case AINT:
				t2 = APINT;
				break;
			case AREAL:
				t2 = APREAL;
				break;
			case AVIRT:
				t2 = APREF;
				break;
			}
			M[a1] = (M[t1 + 2] + M[t1]) / t2 - 1;
		}
		else
			errsignal(RTEREFTN);
		break;
	/* LGETTYPE */
	case 40:
		typep(M[a2], a3, &virt1.addr, &virt1.mark);
		storevirt(virt1, a1);
		break;
	/* LCOPY */
	case 41:
		loadvirt(virt2, a2);
		copy(&virt2, &virt1);
		storevirt(virt1, a1);
		break;
	/* LNOT */
	case 42:
		M[a1] = ~M[a2];
		break;
	/* LRCVAVIRT */	/* recover virtual address from ah */
	case 43:
		virt1.addr = M[a2];
		virt1.mark = M[virt1.addr + 1];
		storevirt(virt1, a1);
		break;
	/* LVIRTDOT */
	case 44:
	case 45:
		M[a1] = M[prototype[M[temporary]]->virtlist + a2];
		break;
	/* LADDRPH */
	case 46:
	/* LADDRPH2 */
	case 47:
		/* fprintf(stderr, "co ja tu robie?"); */
		loadvirt(virt2, a2);
		if (!member(&virt2, &M[a1]))
			errsignal(RTEREFTN);
		break;
	/* LIABS */
	case 48:
		t2 = M[a2];
		M[a1] = absolute(t2);
		break;
	/* LINEG */
	case 49:
		M[a1] = -M[a2];
		break;
	/* LRABS */
	case 50:
		r = MR(a2);
		if(r < (real)0.0)
			r = (real)0.0 - r;
		MR(a1) = r;
		break;
	/* LRNEG */
	case 51:
		MR(a1) = -MR(a2);
		break;
	/* LPARAMADDR */
	case 52:
	 	t2 = M[a2];
		M[a1] = t2 + M[prototype[M[t2 + PROTNUM]]->parlist + a3];
		break;
	/* LLOADT */
	case 54:
		/* offset */
		t1 = M[ic++];
		/* object address */
		t2 = t1 + loadt(M[M[a2]], a3);
		loadvirt(virt1, t2);
		storevirt(virt1, a1);
		break;
	/* LIS */
	case 55:
		loadvirt(virt2, a2);
		M[a1] = lbool(is(&virt2, a3));
		break;
	/* LIN */
	case 56:
		loadvirt(virt2, a2);
		M[a1] = lbool(inl(&virt2, a3));
		break;
	/* LQUA */
	case 57:
		loadvirt(virt2, a2);
		if (member(&virt2, &M[a1]))
			qua(&virt2, a3);
		else
			errsignal(RTEREFTN);
		break;
	/* LIFIX */
	case 58:
		M[a1] = (word)(MR(a2));
		break;
	/* LFLOAT */
	case 59:
		MR(a1) = (real)(M[a2]);
		break;
	/* LIMOVE */
	case 60:
		M[a1] = M[a2];
		break;
	/* LVMOVE */
	case 61:
		loadvirt(virt1, a2);
		storevirt(virt1, a1);
		break;
	/* LRMOVE
	WARNING: these areas may overlap! */
	case 62:
		r = MR(a2);
		MR(a1) = r;
		break;
	/* LFPMOVE
	WARNING: these areas may overlap!*/
	case 63:
		/* MACHINE DEPENDENT */
		loadvirt(virt1, a2);
		t1 = M[a2 + 2];
		storevirt(virt1, a1);
		M[a1 + 2] = t1;
		break;
	/* LEQNONE */
	case 82:
		M[a1] = lbool(M[a2 + 1] != M[M[a2] + 1]);
		break;
	/* LNENONE */
	case 83:
		M[a1] = lbool(M[a2+1] == M[M[a2] + 1]);
		break;
	/* LMDFTYPE */
	/* modify the formal type */
	case 87:
		loadvirt(virt1, a2);
		/* number of "arrayof" */
		virt1.addr += a3;
		storevirt(virt1, a1);
		break;
	/* LOR */
	case 100:
		M[a1] = M[a2] | M[a3];
		break;
	/* LAND */
	case 101:
		M[a1] = M[a2] & M[a3];
		break;
	/* LARRAY */
	case 102:
	case 103:
	case 104:
		loadvirt(virt2, a2);
		if (member(&virt2, &t2)) {
			/* index-lower+3 */
			t1 = M[a3] - M[t2 + 2];
			if (t1 < 3 || t1 >= M[t2])
				errsignal(RTEINVIN);
			else
				M[a1] = t2+t1;
		}
		else
			errsignal(RTEREFTN);
		break;
		
	/* LFARRAY
	without any tests */
	case 105:
		/* physical address */
		t1 = M[M[a2]];
		M[a1] = t1 + M[a3] - M[t1 + 2];
		break;
	/* LIEQUAL */
	case 106:
		M[a1] = lbool(M[a2] == M[a3]);
		break;
	/* LINEQUAL */
	case 107:
		M[a1] = lbool(M[a2] != M[a3]);
		break;
	/* LILT */
	case 108:
		M[a1] = lbool(M[a2] < M[a3]);
		break;
	/* LILE */
	case 109:
		M[a1] = lbool(M[a2] <= M[a3]);
		break;
	/* LIGT */
	case 110:
		M[a1] = lbool(M[a2] > M[a3]);
		break;
	/* LIGE */
	case 111:
		M[a1] = lbool(M[a2] >= M[a3]);
		break;
	/* LCOMBINE */
	case 112:
		loadvirt(virt2, a2);
		t1 = M[a3];
		storevirt(virt2, a1);
		M[a1 + APREF] = t1;
		break;
	/* LIADD */
	case 113:
		M[a1] = M[a2] + M[a3];
		break;
	/* LISUB */
	case 114:
		M[a1] = M[a2]-M[a3];
		break;
	/* LIMULT */
	case 115:
		M[a1] = M[a2] * M[a3];
		break;
	/* LSHIFT */
	case 116:
		M[a1] = shift(M[a2], M[a3]);
		break;
		
	case 117:/* LIDIVE */
		if (M[a3] == 0)
			errsignal(RTEDIVBZ);
		else
			M[a1] = M[a2] / M[a3];
		break;
	/* LIMODE */
	case 118:
		if (M[a3] == 0)
			errsignal(RTEDIVBZ);
		else
			M[a1] = M[a2] % M[a3];
		break;
	/* LRADD */
	case 119:
		MR(a1) = MR(a2) + MR(a3);
		break;
	/* LRSUB */
	case 120:
		MR(a1) = MR(a2) - MR(a3);
		break;
	/* LRMULT */
	case 121:
		MR(a1) = MR(a2) * MR(a3);
		break;
	/* LRDIVE */
	case 122:
		if (MR(a3) == (real)0.0)
			errsignal(RTEDIVBZ);
		else
			MR(a1) = MR(a2) / MR(a3);
		break;
	/* LEQREF */
	case 123:
		loadvirt(virt2, a2);
		loadvirt(virt3, a3);
		if (member(&virt2, &t1))
			M[a1] = lbool(member(&virt3, &t2) && t1 == t2);
		else
			M[a1] = lbool(!member(&virt3, &t2));
		break;
	/* LNEREF */
	case 124:
		loadvirt(virt2, a2);
		loadvirt(virt3, a3);
		if (member(&virt2, &t1))
			M[a1] = lbool(!member(&virt3, &t2) || t1 != t2);
		else
			M[a1] = lbool(member(&virt3, &t2));
		break;
	/* LREQ */
	case 125:
		M[a1] = lbool(MR(a2) == MR(a3));
		break;
	/* LRNE */
	case 126:
		M[a1] = lbool(MR(a2) != MR(a3));
		break;
	/* LRLT */
	case 127:
		M[a1] = lbool(MR(a2) < MR(a3));
		break;
	/* LRLE */
	case 128:
		M[a1] = lbool(MR(a2) <= MR(a3));
		break;
	/* LRGT */
	case 129:
		M[a1] = lbool(MR(a2) > MR(a3));
		break;
	/* LRGE */
	case 130:
		M[a1] = lbool(MR(a2) >= MR(a3));
		break;
	/* LXOR */
	case 131:
		M[a1] = M[a2] ^ M[a3];
		break;
	/* LCALLPROCSTAND */
	case 132:
#if USE_ALARM
		/* reschedule forced so alarm may be switched off */
		alarm(0);
#endif
		reschedule = TRUE;
		standard(a1);
		break;
	/* LKILL */
	case 143:
		loadvirt(virt1, a1);
		disp(&virt1);
		break;
	/* LHEADS */
	case 144:
		loadvirt(virt1, a1);
		heads(&virt1, a2);
		break;
	/* LIPARINP */
	case 145:
		param[a3].xword = M[a1];
		break;
	/* LGKILL */
	case 146:
		loadvirt(virt1, a1);
		gkill(&virt1);
		break;
	/* LVPARINP */
	case 147:
		loadvirt(param[a3].xvirt, a1);
		break;
	/* LRPARINP */
	case 148:
		param[a3].xreal = MR(a1);
		break;
	/* LQUATEST */
	case 149:
		loadvirt(virt1, a1);
		qua(&virt1, a2);
		break;
	/* LSTYPE */
	case 150:
		loadvirt(virt1, a1);
		typref(&virt1, a2);
		break;
	/* LIFFALSE */
	case 151:
		if (M[a1] == LFALSE)
			ic = a2;
		break;
	/* LIFTRUE */
	case 152:
		if (M[a1] == LTRUE)
			ic = a2;
		break;
	/* LGO */
	case 159:
		go(M[a2], M[a1]);
		break;
	/* LGOLOCAL */
	case 160:
		goloc(M[a2], M[a1]);
		break;
	/* LDTYPE */
	case 170:
		/* left side type */
		loadvirt(virt1, a1);
		loadvirt(virt2, a2);
		/* right side type */
		loadvirt(virt3, a3);
		typed(virt1.addr, virt1.mark, virt3.addr, virt3.mark, &virt2);
		break;
	/* LTERMINATE */
	case 172:
		term();
		break;
	/* LWIND */
	case 173:
		wind();
		break;
	/* LBLOCK2 */
	case 174:
		goloc(thisp->blck1, thisp->blck2);
		break;
	/* LBLOCK3 */
	case 176:
		disp(&thisp->backobj);
		break;
	/* LTRACE */
	case 177:
		trace(a1);
		break;
	/* LINNER */
	case 178:
		inner(a1);
		break;
	/* LBACKHD */
	case 180:
		backhd(&thisp->backobj, &M[temporary]);
		break;
	/* LJUMP */
	case 182:
		ic = a1;
		break;
	/* LBLOCK1 */
	case 186:
		openobj(a1, &thisp->blck1, &thisp->blck2);
		break;
	/* LDETACH */
	case 187:
		detach();
		break;
	/* LATTACH */
	case 188:
		loadvirt(virt1, a1);
		attach(&virt1);
		break;
	/* LBACKBL */
	case 191:
		backbl(&thisp->backobj, &M[temporary]);
		break;
	/* LBACKPR */
	case 192:
		/* backpr(&thisp->backobj, &M[temporary]); */
		back(&thisp->backobj, &M[temporary], (word) 0);
		break;
	/* LBACK */
	case 193:
		back(&thisp->backobj, &M[temporary], (word) 0);
		break;
	/* LFIN */
	case 194:
		fin(ic - APOPCODE, &thisp->backobj, &M[temporary]);
		break;
	/* LCASE */
	case 195:
		/* a2 = address of case description : */
		/* minimal value, number of branches, */
		/* remaining branches followed by "otherwise" code */
		/* in 0..number of branches-1 */
		t1 = M[a1] - M[a2];
		if (t1 < 0 || t1 >= M[a2 + 1])
			/* otherwise */
			ic = a2 + 2 + M[a2 + 1];
		else
			/* indirect jump */
			ic = M[a2 + 2 + t1];
		break;
	/* LRESUME */
	case 220:
		loadvirt(virt1, a1);
		resume(&virt1);
		break;
	/* LSTOP */
	case 221:
		passivate(STOPPED);
		break;
	/* LKILLTEMP */
	case 222:
		disp(&thisp->template);
		break;
	/* LENABLE */
	case 223:
		for (i = 0; i < a1; i++)
			enable(thispix, virtprot(M[ic++]));
		evaluaterpc(thispix);
		break;
	/* LDISABLE */
	case 224:
		for (i = 0; i < a1; i++)
			disable(thispix, virtprot(M[ic++]));
		break;
	/* LACCEPT1 */
	case 225:
		rpc_accept(a1);
		break;
	/* LACCEPT2 */
	case 226:
		popmask(thispix);
		rpc3();
		break;
	/* LBACKRPC */
	case 227:
		back(&thisp->backobj, &M[temporary], a1);
		break;
	/* LASKPROT */
	case 228:
		loadvirt(virt1, a1);
		askprot(&virt1);
		break;
	/* LSTEP */
	case 240:
		if (M[a1] < 0)
			errsignal(RTENEGST);
		break;

	default:
		break;
	}
}


