/*     Loglan82 Compiler&Interpreter
     Copyright (C) 1993 Institute of Informatics, University of Warsaw
     Copyright (C)  1993, 1994 LITA, Pau
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
	     You should have received a copy of the GNU General Public License
	     along with this program; if not, write to the Free Software
	     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 contacts:  Andrzej.Salwicki@univ-pau.fr

or             Andrzej Salwicki
		LITA   Departement d'Informatique
		Universite de Pau
		Avenue de l'Universite
		64000 Pau   FRANCE
		 tel.  ++33 59923154    fax. ++33 59841696

=======================================================================
*/

#include "depend.h"
#include "genint.h"
#include "int.h"
#include "process.h"
#include "intproto.h"

#include <stdio.h>

/* File I/O routines */


/**
 * Load parameters of current file
 * @parameter expected status of file
 * @parameter file type
 * @parameter file object address
 * @parameter file stream pointer
 */
void loadfile(word status, word *ftype, word *am, FILE **fp)
{
	word s;
	virtaddr virt;

	loadvirt(virt, currfile);
	/* file object exists */
	if (member(&virt, am)) {
		/* check status */
		s = M[*am + FSTAT];
		if (status != s && status != UNKNOWN)
			errsignal(RTEILLIO);
		*ftype = M[*am + FTYPE];
		*fp = MF(*am + FFILE);
	} else {
		/* file not opened yet */
		errsignal(RTEREFTN);
	}
}


/**
 * Open file object
 * @parameter TRUE iff file is temporary
 * @parameter file type
 * @parameter file name
 * @parameter output virtual address
 * @parameter output physical address
 */
void genfileobj(bool ftemp, word ftyp, char *fnam, virtaddr *virt, word *am)
{
	word t1;

	/* generate file object */
	request((word) APFILE, &t1, am);
	virt->addr = t1;
	virt->mark = M[t1 + 1];
	M[*am + PROTNUM] = FILEOBJECT;
	M[*am + FSTAT] = UNKNOWN;
	M[*am + FTEMP] = lbool(ftemp);
	M[*am + FTYPE] = ftyp;
	MN(*am + FNAME) = fnam;
}

/* Prepare file for reading */
void reset(word am)
{
	FILE *fp;

	/* first close file if opened */
	if (M[am + FSTAT] != UNKNOWN)
		if (fclose(MF(am + FFILE)))
			errsignal(RTEIOERR);

	switch ((int) M[am + FTYPE]) {
	/* open text file for reading */
	case TEXTF:
		fp = fopen(MN(am + FNAME), "r");
		M[am + FSTAT] = READING;
		break;

	/* open binary file for reading */
	case CHARF:
	case INTF:
	case REALF:
		fp = fopen(MN(am + FNAME), BINARYREAD);
		M[am + FSTAT] = READING;
		break;

	/* open existing file for update */
	case DIRECT:
		fp = fopen(MN(am + FNAME), DIRECTOLD);
		M[am + FSTAT] = UPDATING;
		break;
	}
	if (fp == NULL) {
		M[am + FSTAT] = UNKNOWN;
		errsignal(RTECNTOP);
	}
	/* store stream pointer */
	MF(am + FFILE) = fp;
}

/**
 * Prepare file for writing
 */
void rewrite(word am)
{
	FILE *fp;

	/* first close file if opened */
	if (M[ am+FSTAT ] != UNKNOWN)
		if (fclose(MF(am + FFILE)))
			errsignal(RTEIOERR);

	switch ((int) M[am + FTYPE]) {
	/* open text file for writing */
	case TEXTF:
		fp = fopen(MN(am + FNAME), "w");
		M[am + FSTAT] = WRITING;
		break;

	/* open binary file for writing */
	case CHARF:
	case INTF:
	case REALF:
		fp = fopen(MN(am + FNAME), BINARYWRITE);
		M[am + FSTAT] = WRITING;
		break;

	/* create new file for update */
	case DIRECT:
		fp = fopen(MN(am+FNAME), DIRECTNEW);
		M[am + FSTAT] = UPDATING;
		break;
	}

	if (fp == NULL) {
		M[am + FSTAT] = UNKNOWN;
		errsignal(RTECNTOP);
	}
	/* store stream pointer */
	MF(am + FFILE) = fp;
}

/**
 * Delete file
 */
void delete(virtaddr *virt)
{
	word am;

	if (member(virt, &am)) {
		/* first close file if opened */
		if (M[am + FSTAT ] != UNKNOWN)
			if (fclose(MF(am+FFILE))) errsignal(RTEIOERR);

		/* delete file */
		if (unlink(MN(am+FNAME)))
			errsignal(RTEIOERR);
		/* free memory used by file name */
		free(MN(am+FNAME));
		/* and kill file object */
		disp(virt);
	} else {
		errsignal(RTEREFTN);
	}
}

/**
 * Generate temporary file name
 */
char *tempfilename()
{
	char *cp;
	static int tempcnt = 0;

	cp = ballocate(10);
	if (cp == NULL)
		errsignal(RTEMEMOV);

	sprintf(cp, "LOG%05d", tempcnt++);
	return cp;
}

/**
 * Test for end of file
 */
bool testeof(FILE *fp)
{
	int ch;

	ch = getc(fp);
	ungetc(ch, fp);
	return (ch == EOF);
}

/**
 * Test for end of line
 */
bool testeoln(FILE *fp)
{
	int ch;

	ch = getc(fp);
	ungetc(ch, fp);
	return (ch == '\n');
}

/**
 * Skip to end of line
 */
void readln(FILE *fp)
{
	int ch, st;
	G_MESSAGE msg;
	if (fp == stdin) {
		read_line();
	}
	else {
		while (ch != '\n' && ch != EOF) {
			ch = getc(fp);
		}
	}
}

static char str[10];
/**
 * Read integer
 */
word readint(FILE *fp)
{
	long i = 0L;
	int j = 0, c = 0;
	int bool = 0;


	while(c < '0' || c > '9') {
		if(c == '-')
			bool = 1;
		else
			bool = 0;

		c=fgetc(fp);

		if(c == EOF) {
			errsignal(RTEBADFM);
			goto END;
		}
	}

	do {
		i = 10*i + (c - '0');
		j++;
		c = fgetc(fp);
	} while(c >= '0' && c <= '9');

	if(c != EOF)
		ungetc(c, fp);
	if (j == 0 )
		errsignal(RTEBADFM);

END:
	if (bool)
		return -i;
	else
		return i;
}

/**
 * Read real
 */
double readreal(FILE *fp)
{
	double r;

	if (fscanf(fp, "%lf", &r) != 1)
		errsignal(RTEBADFM);

	return r;
}

/**
 * Write integer
 */
void writeint(word n, word field, FILE *fp)
{
	static char format[32];

	sprintf(format,"%*ld",(int)field, (long)n);

	if (fp == stdout)
		write_str(format);
	else if (fprintf(fp, "%*ld", (int)field, (long) n) == 0)
		errsignal(RTEIOERR);
}

/**
 * Write real
 */
void writereal(double r, word field1, word field2, FILE *fp)
{
	char format[32];

	sprintf(format, "%*.*lf", (int) field1, (int) field2,r);
	if (fp == stdout)
		write_str(format);
	else if (fprintf(fp,"%*.*lf", (int)field1, (int)field2, r) == 0)
		errsignal(RTEIOERR);
}

/**
 * Write string
 */
void writestring(word offset, word field, FILE *fp)
{
	word len, addr;
	int i;
	char *cp;
	char s[256];

	addr = strings + offset;
	len = M[addr];
	/* pointer to first char of string */
	cp = (char *) &M[ addr+1 ];
	if (fp == stdout) {
		for(i = 0; i < len; i++)
			s[i] = *cp++;
		s[len] = '\0';
		write_str(s);
	} else {
		while (len-- > 0 && field-- != 0) {
			if (putc(*cp++, fp) == EOF)
				errsignal(RTEIOERR);
		}
	}
}

/**
 * Perform direct access read/write
 * @parameter buffer array
 * @parameter number of bytes to transfer
 * @parameter fread() or fwrite()
 * @parameter stream pointer
 */
#ifndef NO_PROTOTYPES
word directio(virtaddr *buf, word len, int (*action)(char *,int, int, FILE *),
								FILE *fp)
#else
word directio(virtaddr *buf, word len, int (*action)(), FILE *fp)
#endif
{
	word am, t1, result;
	int n;

	/* file not none */
	if (member(buf, &am)) {
		/* seek to current position required*/
		if (fseek(fp, 0L, 1))
			errsignal(RTEIOERR);
		
		/* check appetite */
		len = min(len, (M[ am ]-3)*sizeof(word));
		/* number of bytes transfered */
		result = 0;
		/* address in memory for transfer */
		t1 = am+3;
		/* transfer full blocks */
		while (len >= IOBLOCK) {
			n = (*action)((char *) &M[t1], 1, IOBLOCK, fp);
			result += n;
			if (n != IOBLOCK)
				return result;
			len -= IOBLOCK;
			t1 += IOBLOCK / sizeof(word);
		}
		/* transfer last unfilled block */
		if (len > 0) {
			n = (*action)((char *) &M[t1], 1, (int) len, fp);
			result += n;
		}
		return(result);
	}
	else {
		errsignal(RTEREFTN);
	}
}

