#include "depend.h"
#include "genint.h"
#include "int.h"
#include "process.h"
#include "intproto.h"

/**
 * @file
 */

/**
 * \defgroup Variables common with generator :
 * @{
 */
 
/**
 * prototypes
 */
protdescr *prototype[MAXPROT + 1];

/**
 * address of primitive types descriptions
 */
word ipradr;

/**
 * address of global temporary variables
 */
word temporary;

/** 
 * base for string constants
 */
word strings;

/**
 * the last used prototype number
 */
word lastprot;

/**
 * first free cell in M
 */
word freem;

/**
 * current file virtual address
 */
word currfile = 2;
/** @} */

/**
 * Interpreter own variables :
 * @{
 **/
 
/**
 * main memory for code and data
 */
memory M;

/**
 * for comunication with standard procs
 */
union value *param;
/** @} */

/**
 * offset conversion table for compactification
 */
int offset[] = { DUMMY, SL, DL, CL, CHD, VIRTSC };

/**
 * signal to number conversion table 
 * -1 stands for an unrecoverable error which cannot be serviced by handler
 */
int scot[] = {
	/** RTESLCOF */
	20,
	/** RTEUNSTP */
	20,
	/** RTEILLAT */
	20,
	/** RTEILLDT */
	20,
	/** RTECORTM */
	20,
	/** RTECORAC */
	20,
	/** RTEINVIN */
	23,
	/** RTEILLAB */
	23,
	/** RTEINCQA */
	21,
	/** RTEINCAS */
	24,
	/** RTEFTPMS */
	20,
	/** RTEILLKL */
	20,
	/** RTEILLCP */
	20,
	/** RTEINCHS */
	24,
	/** RTEHNDNF */
	-1,
	/** RTEMEMOV */
	22,
	/** RTEFHTLG */
	22,
	/** RTEILLRT */
	-1,
	/** RTEREFTN */
	21,
	/** RTEDIVBZ */
	01,
	/** RTESYSER */
	02,
	/** RTEILLIO */
	02,
	/** RTEIOERR */
	02,
	/** RTECNTOP */
	02,
	/** RTEBADFM */
	02,
	/** RTEILLRS */
	20,
	/** RTETMPRC */
	02,
	/** RTEINVND */
	02,
	/** RTENEGST */
	23,
	/** RTENONGL */
	-1
};

/**
 * Primitive type appetites for moveparams() :
 */
int primapet[] = {
	/** INTEGER */
	APINT,
	/** REAL */
	APREAL,
	/** BOOLEAN */
	APINT,
	/** CHAR */
	APINT,
	/** COROUTINE */
	APREF,
	/** PROCESS */
	APREF,
	/** STRING */
	APINT
};

/**
 * instruction counter
 */
word ic;

/**
 * previous ic for redecoding after compact.
 */
word lastic;

/**
 * opcode of L-code instruction
 */
int opcode;

/**
 * arguments of L-code instructions
 */
word a1, a2, a3;

/**
 * \defgroup kernel
 * \brief variables for the running system
 */
/** @{ */
/**
 * code and data memory size
 */
word memorysize = DEFMEMSIZE;

/**
 * pointers to current object
 */
word c1, c2;

/**
 * DISPLAY offset in process object
 */
word dispoff;
/**
 * indirect DISPLAY offset in process object
 */
word disp2off;

/**
 * DISPLAY address - physical addresses
 */
word display;
/**
 * DISPLAY address - indirect addresses
 */
word display2;

/**
 * main block object
 */
word mainprog;

/**
 * offset of variable main
 **/
word mnoff;

/** @} */

/**
 * default: no compactification message
 */
bool infmode = FALSE;

/**
 * TRUE if trace is printed
 */
bool debug = FALSE;

/**
 * output file for trace
 */
FILE *tracefile;

/**
 * for continue execution
 */
jmp_buf contenv;
