#include        "queue.h"


/**
 * @defgroup
 * @brief Process management definitions
 */
/** @{ */

/** maximum number of processes on one node */
#define MAXPROCESS       64

/** maximum number of waiting messages */
#define MAXMSGQUEUE      16

/** message length defined by me (PS) */
#define MSGLENGTH       256
/** @} */

/**
 * @defgroup
 * @brief Process states
 */
/** @{ */
/* during generation of process object */
#define GENERATING      0

/** non-active process (suspended by STOP) */
#define STOPPED         1

/** active process (ready to execute) */
#define EXECUTING       2

/** waiting for NEW of another process */
#define WAITFORNEW      3

/** waiting for remote procedure call */
#define WAITFORRPC      4

/** during execution of ACCEPT statement */
#define ACCEPTING       5

/** waiting for process prototype */
#define WAITASKPRO      6
/** @} */

/** Process descriptor */
typedef struct
{
	/* TRUE if in use by some process */
	bool used;

	/* process mark for proper detecting of none */
	word mark;

	/* process state */
	int status;

	/* process prototype number */
	word prot;

	/* pointer to memory array */
	memory M;

	union value param[MAXPARAM];

	/* instruction counter */
	word ic;

	/* trace line number */
	word trlnumber;

	/* first word of object area */
	word lower;

	/* last word in memory */
	word upper;

	/* last word used by objects */
	word lastused;

	/* first word used by dictionary */
	word lastitem;

	/* head of free dictionary item list */
	word freeitem;

	/* head of killed object list for size > 2 */
	word headk;

	/* head of killed object list for size = 2 */
	word headk2;

	/* am of process object */
	word prochead;

	/* process object virtual address */
	virtaddr procref;

	/* remote process or procedure template */
	virtaddr template;

	/* pointers to current object */
	word c1, c2;

	/* adress of object just left */
	virtaddr backobj;

	/* used for LBLOCK1, LBLOCK2, LBLOCK3 */
	word blck1, blck2;

	/* queue of messages for this process */
	queue msgqueue;

	/* queue of disabled RPC messages */
	queue rpcwait;

	/* stack of set of enabled remote procedures */
	stack rpcmask;

	/* next allocate will forace compact */
	bool force_compactification;

	/* table of pointers to processes in process */
	word *hash;
	word hash_size;
} procdescr;


/**
 * Message type:
 * \defgroup MessageType Message type
 * @{
 */

/** error signal */
#define ERRSIG   0
/** resume request */
#define RESUME   1
/** create new process request */
#define CREATE   2
/** create process acknowledge */
#define CREACK   3
/** kill process */
#define KILLPR   4
/** remote procedure call request */
#define RPCALL   5
/** remote procedure return */
#define RPCACK   6
/** ask for process prototype */
#define ASKPRO   7
/** answer with process prototype */
#define PROACK   8

/** @} */

typedef struct {
	word node;
	word pix;
	word mark;
} procaddr;

struct ctrlmsg
{
	/** address of the sender and */
	procaddr sender;

	/** receiver of the message */
	procaddr receiver;

	/** message type */
	int type;

	/** prototype or error signal number */
	int par;
};

#define MAXPROCPAR      (MSGLENGTH-sizeof(struct ctrlmsg))

typedef struct
{
	struct ctrlmsg control;
	char params[MAXPROCPAR];
} message;

/* Direction of copying of parameters (for moveparams()): */

#define LOADPAR         0
#define SAVEPAR         1

typedef char *mask;

/** process descriptor table */
extern procdescr process[];

/** pointer to current process descriptor */
extern procdescr *thisp;

/** current process index */
extern word thispix;

/** Round-Robin queue of ready processes */
extern queue ready;

/** TRUE if operating in D-Link network */
extern bool network;

/** queue of waiting messages */
extern message globmsgqueue[];

/** number of waiting messages */
extern int msgready;

/** pointers to message queue */
extern int msghead, msgtail;

/** this machine node number */
extern word ournode;

/** console node number */
extern word console;

/** TRUE if remote node */
extern bool remote;

/** TRUE if rescheduling is mandatory */
extern bool reschedule;

#if OS2
/** pointer to Global Info Segment */
extern PGINFOSEG ginf;
#endif



#ifndef NO_PROTOTYPES
void obj2mess(word *,virtaddr *,procaddr*);
void mess2obj(procdescr *,procaddr *,virtaddr*);
bool isprocess(virtaddr *);
void hash_find(procaddr *,virtaddr *);
void hash_create(procdescr *,int);
void hash_set(procaddr *,word);
#else
void obj2mess();
void mess2obj();
bool isprocess();
void hash_find();
void hash_create();
void hash_set();
#endif

