     /* Loglan82 Compiler&Interpreter
     Copyright (C) 1981-1993 Institute of Informatics, University of Warsaw
     Copyright (C)  1993, 1994 LITA, Pau
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
             You should have received a copy of the GNU General Public License
             along with this program; if not, write to the Free Software
             Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 contacts:  Andrzej.Salwicki@univ-pau.fr

or             Andrzej Salwicki
                LITA   Departement d'Informatique
                Universite de Pau
                Avenue de l'Universite
                64000 Pau   FRANCE
		 tel.  ++33 59923154	fax. ++33 59841696

=======================================================================
*/

#include "depend.h"
#include "genint.h"
#include "int.h"
#include "process.h"
#include "intproto.h"

/**
 * @file
 */

/**
 * Initialize memory structures for objects, main object and a few goodies more.
 */
void runsys()
{
	word apt, i;
	procaddr father;

	/* initialize process descriptors */
	for (i = 0; i < MAXPROCESS; i++) {
		/* not used */
		process[i].used = FALSE;
		/* initial mark for processes */
		process[i].mark = -1;
		/* memory not allocated */
		process[i].M = NULL;
		process[i].hash = NULL;
	}
	/* always contains code */
	process[0].M = M;
	/* DISPLAY offset in process object */
	dispoff = VIRTSC - (lastprot + 1);
	/* indirect DISPLAY offset */
	disp2off = dispoff - (lastprot + 1);
	/* initialize Round-Robin queue */
	ready = qinit();
	/* init pseudo-random no. generator */
	ranset();

#if OS2
	{
		SEL gsel, lsel;
		DosGetInfoSeg(&gsel, &lsel);
		ginf = MAKEPGINFOSEG(gsel);
	}
#endif

	/* create main process */
	if (!remote) {
		/* dummy DL for generated process */
		father.node = 0;
		father.pix  = 0;
		father.mark = 0;
		/* current process index */
		thispix = 0;
		/* current process descr pointer */
		thisp = &process[thispix];
		initprocess((word) 0, (word) MAINBLOCK, &father);
		/* am of main */
		mainprog = thisp->prochead;
		/* pointers to current object */
		c1 = thisp->c1;
		c2 = thisp->c2;
		/* instruction counter */
		ic = thisp->ic;
		/* parameter vector */
		param = thisp->param;
		/* LWA+1 of main */
		apt = mainprog + M[mainprog];
		/* DISPLAY in main */
		display = apt + dispoff;
		/* indirect DISPLAY in main */
		display2 = apt + disp2off;
		/* offset of variable mainprog */
		mnoff = 2;
		/* init variable main */
		storevirt(thisp->procref, mainprog + mnoff);
		/* flag main included in SL chain */
		M[apt + STATSL]++;
		thisp->status = STOPPED;
		/* activate main process */
		activate(thispix);
	}
	/* remote */
	else {
		/* a dirty trick: set junk current */
		thispix = 0;
		/*  process for first transfer() */
		thisp =  &process[thispix];
		/* (must save 'context' somewhere) */
	}
#if DLINK
	net_attention();
#endif
}

/**
 * Initialize process descriptor
 */
void initprocess(word pix, word prot, procaddr *father)
{
	procdescr *p;
	protdescr *ptr;
	word i, j, ah, am, apt;

#ifdef RPCDBG
	fprintf(stderr, "new process(n,p,m) (%d,%d,%d)", 0, pix,
							process[pix].mark);
	fprintf(stderr, " from (%d,%d,%d)\n", father->node, father->pix,
								father->mark);
#endif

	p = &process[pix];

#ifdef OBJECTADDR
	hash_create(p, 119);
#endif
	/* process descriptor is used */
	p->used = TRUE;
	/* prototype number */
	p->prot = prot;
	/* null list of free dictionary items */
	p->freeitem = 0;
	/* highest memory address */
	p->upper = memorysize - 1;
	/* lowest address for data */
	p->lower = freem;
	/* head of killed objects list */
	p->headk = p->lower;
	/* maximum appetite sentinel */
	p->M[ p->headk ] = MAXAPPT;
	p->headk2 = 0;
	/* dict. item for process itself */
	ah = p->upper - 1;
	/* first word used by dictionary */
	p->lastitem = ah;
	ptr = prototype[prot];
	if (p->upper - p->lower - ptr->appetite < 512)
		if (prot == MAINBLOCK)
			abend("Memory size too small (use /m option)\n");
		else errsignal(RTEMEMOV);

    /* generate process object */
	p->lastused = p->lower + ptr->appetite;
	am = p->lower + 1;
	p->M[am] = ptr->appetite;
	p->M[am + PROTNUM] = prot;
	for (i = PROTNUM + 1; i < ptr->appetite; i++)
		p->M[am + i] = 0;
	p->M[ah] = am;
	p->M[ah + 1] = 0;
	p->prochead = am;
	p->procref.addr = ah;
	p->procref.mark = 0;
	/* initialize current object ptrs */
	p->c1 = am;
	p->c2 = am + ptr->span;
	apt = am + ptr->appetite;
	/* initialize coroutine head ptr */
	p->M[apt + CHD] = ah;
	p->M[apt + CHD + 1] = 0;
	/* dummy SL for process */
	p->M[apt + SL] = DUMMY;
	/* absolute none */
	p->M[1] = 1;
	/* initialize DISPLAY */
	for (i = MAINBLOCK; i <= lastprot; i++) {
		/* dummmy entry for MAIN */
		p->M[apt + dispoff + i] = 0;
	}
	p->M[apt + disp2off + MAINBLOCK] = DUMMY;
	/* set DISPLAY entries for process */
	j = ptr->preflist;

	for (i = j + ptr->lthpreflist - 1; i >= j; i--) {
		/* physical address */
		p->M[apt + dispoff + M[i]] = am;
		/* indirect address */
		p->M[apt + disp2off + M[i]] = ah;
	}

	{
		virtaddr v;
		mess2obj(p, father, &v);
		p->M[apt + DL] = v.addr;
		p->M[apt + DL + 1] = v.mark;
	}
	p->msgqueue = qinit();
	p->rpcwait = qinit();
	p->rpcmask = sinit();
	/* initialy all RPCs are disabled */
	pushmask(pix);
	/* trace line number */
	p->trlnumber = 0;
	/* search for executable prefix */
	i = ptr->preflist;
	while (prototype[p->M[i]]->kind == RECORD)
		i++;

	/* first instruction address */
	p->ic = prototype[M[i]]->codeaddr;
	p->force_compactification = FALSE;
}


/**
 * 
 */
bool member(virtaddr *virt, word *am)
{
	*am = M[virt->addr];
/*
	if (virt->mark == M[virt->addr + 1])
		fprintf(stderr, "Yes");
	else {
		fprintf(stderr, "No");
	}
*/
	return (virt->mark == M[virt->addr + 1]);
}

/**
 * Update DISPLAY
 */
void update(word am, word ah)
{
	word t1, t2, t3, t4, t5, t6;
	protdescr *ptr;

	while (TRUE) {
		t1 = am + M[am];
		/* flag object included in SL */
		M[t1 + STATSL]++;
		ptr = prototype[M[am + PROTNUM]];
		t2 = ptr->preflist;
		t3 = t2 + ptr->lthpreflist - 1;
		for (t4 = t3; t4 >= t2; t4--) {
			t6 = M[t4];
			t5 = display + t6;
			/* entry to be updated */
			if (M[ t5 ] == 0) {
				M[t5] = am;
				M[display2 + t6] = ah;
			}
		}
		ah = M[t1 + SL];
		if (ah == DUMMY)
			break;

		if (M[ah + 1] != M[t1 + SL + 1])
			errsignal(RTESLCOF);

		am = M[ah];
	}
}

/**
 * Loosen DISPLAY
 */
void loosen()
{
	word t1, t2, t3;
	protdescr *ptr;

	t1 = c1;
	while (TRUE) {
		ptr = prototype[M[t1 + PROTNUM]];
		t2 = ptr->preflist;
		for (t3 = t2 + ptr->lthpreflist - 1; t3 >= t2; t3--)
			M[display + M[t3]] = 0;
		t3 = t1 + M[t1];
		/* flag object removed from SL */
		M[t3 + STATSL]--;
		/* ah of SL */
		t1 = M[t3 + SL];
		/* still not main */
		if (t1 == DUMMY)
			break;
		/* am of SL */
		t1 = M[t1];
	}
}

/**
 * To count trace messages in line
 */
static int tracecnt = 0;

/**
 * Trace the program if debug mode
 */
void trace(word lineno)
{
	thisp->trlnumber = lineno;
	if (debug && lineno > 0) {
		tracecnt++;
		/* change line */
		if (tracecnt == MAXTRACNT) {
			tracecnt = 0;
			fprintf(tracefile, "\n");
		}
		fprintf(tracefile, "%6ld", (long) lineno);
	}
	/* check for waiting message */
	trapmsg();
	/* check for RPC message */
	rpc2();
}

/**
 * 
 */
void endrun(int status)
{
	MESSAGE msg;
	G_MESSAGE m;
	int i;

	if (debug)
		fclose(tracefile);

	msg.msg_type = MSG_INT;
	msg.param.pword[0] = INT_EXITING;
	strcpy(msg.param.pstr, ProgName);
	write(internal_sock, &msg, sizeof(MESSAGE));
	m.msg_type = MSG_GRAPH;
	m.param.pword[0] = GRAPH_FREE;
	write(graph_sock, &m, sizeof(G_MESSAGE));
	close(internal_sock);
	close(graph_sock);
	close(net_sock);
	unlink(mygname);
	unlink(mykname);
	unlink(mynname);
	for(i = 0; i < 255; i++)
		if (DirConn[i] != -1)
			close(DirConn[i]);
	exit(status);
}


