     /* Loglan82 Compiler&Interpreter
     Copyright (C) 1981-1993 Institute of Informatics, University of Warsaw
     Copyright (C)  1993, 1994 LITA, Pau
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
             You should have received a copy of the GNU General Public License
             along with this program; if not, write to the Free Software
             Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 contacts:  Andrzej.Salwicki@univ-pau.fr

or             Andrzej Salwicki
                LITA   Departement d'Informatique
                Universite de Pau
                Avenue de l'Universite
                64000 Pau   FRANCE
		 tel.  ++33 59923154	fax. ++33 59841696

=======================================================================
*/

#include "depend.h"
#include "genint.h"
#include "int.h"
#include "process.h"
#include "intproto.h"

#include <math.h>
#include <time.h>

/**
 * @file
 * @brief Call standard procedure
 */

/** Process call to a standard proc. */
void standard(word nrproc)
{
	word t1, t2, t3, t5, t6;
	double r;
	bool absent;
	int ch, n;
	long tim;
	char *cp;
	FILE *fp;
	char s[80];

	absent = FALSE;

#ifdef TRACE
	fprintf(stderr, "standard procedure %d\n", nrproc);
#endif

	switch ((int) nrproc) {
	/* new array */
	case 1:
		newarry(param[1]. xword, param[0].xword, param[2].xword,
					&param[3].xvirt, &param[4].xword);
		break;

	/* rew */
	case 2:
	/* avf */
	case 3:
	/* bsf */
	case 4:
	/* weo */
	case 5:
	/* putrec */
	case 6:
	/* getrec */
	case 7:
	/* ass */
	case 8:
	/* assin */
	case 9:
	/* assout */
	case 10:
		absent = TRUE;
		break;

	/* unpack:function(s:string):arrayof char */
	case 11:
		t1 = strings+param[0].xword+1;
		/* length of the string */
		t6 = M[t1 - 1];
		/* string not null */
		if (t6 > 0) {
			newarry((word) 1, t6, (word)AINT, &param[1].xvirt, &t5);
			t5 += 3;
			cp = (char *) &M[t1];
			while (t6-- > 0) {
				M[t5++] = *cp++;
			}
		} else {
			/* null string */
			param[1].xvirt.addr = 0;
			param[1].xvirt.mark = 0;
		}
		break;
	/* random:function:real */
	case 12:
		param[0].xreal = (real)prandom();
		break;
	/* time:function:integer */
	case 13:
		time(&tim);
		param[0].xword = tim;
		break;
	/* sqrt:function(x:real):real */
	case 14:
		param[1].xreal = (real)sqrt((double) param[0].xreal);
		break;
	/* entier:function(x:real):integer */
	case 15:
		param[1].xword = entier((double) param[0].xreal);
		break;
	/* round:function(x:real):integer */
	case 16:
		param[1].xword = entier((double) (param[0].xreal + 0.5));
		break;
	/* unused */
	case 17:
	/* intrinsic procedure */
	case 18:
		absent = TRUE;
		break;
	/* imin:function(x, y:integer):integer */
	case 19:
		param[2].xword = min(param[0].xword, param[1].xword);
		break;
	/* imax:function(x, y:integer):integer */
	case 20:
		param[2].xword = max(param[0].xword, param[1].xword);
		break;
	/* imin3:function(x, y, z:integer):integer */
	case 21:
		t1 = min(param[0].xword, param[1].xword);
		param[3].xword = min(t1, param[2].xword);
		break;
	/* imax3:function(x, y, z:integer):integer */
	case 22:
		t1 = max(param[0].xword, param[1].xword);
		param[3].xword = max(t1, param[2].xword);
		break;
	/* sin:function(x:real):real */
	case 23:
		param[1].xreal = (real)sin((double) param[0].xreal);
		break;
	/* cos:function(x:real):real */
	case 24:
		param[1].xreal = (real)cos((double) param[0].xreal);
		break;
	/* tan:function(x:real):real */
	case 25:
		r = cos((double) param[0].xreal);
		if (r == 0.0)
			errsignal(RTEDIVBZ);
		param[1].xreal = (real)sin((double) param[0].xreal) / r;
		break;
	/* exp:function(x:real):real */
	case 26:
		param[1].xreal = (real)exp((double) param[0].xreal);
		break;
	/* ln:function(x:real):real */
	case 27:
		param[1].xreal = (real)log((double) param[0].xreal);
		break;
	/* atan:function(x:real):real */
	case 28:
		param[1].xreal = (real)atan((double) param[0].xreal);
		break;
	/* endrun:procedure */
	case 29:
		endrun(0);
		break;
	/* ranset:procedure(x:real) */
	case 30:
		ranset();
		break;
	/* clock */
	case 31:
	/* option */
	case 32:
	/* lock */
	case 33:
	/* unlock */
	case 34:
	/* sched, boy! */
	case 35:
	/* date */
	case 36:
	/* execpar */
	case 37:
	/* test&set */
	case 38:
		absent = TRUE;
		break;
	/* eof */
	case 39:
		param[0].xbool = lbool(testeof(stdin));
		break;
	/* eof(f) */
	case 40:
		loadfile((word) UNKNOWN, &t1, &t2, &fp);
		t3 = M[t2 + FSTAT];
		if (t3 == READING || t3 == UPDATING)
			param[0].xbool = lbool(testeof(fp));
		else
			errsignal(RTEILLIO);
		break;
	/* readln */
	case 41:
		readln(stdin);
		break;
	/* readln(f) */
	case 42:
		loadfile((word) READING, &t1, &t2, &fp);
		if (t1 != TEXTF)
			errsignal(RTEILLIO);
		readln(fp);
		break;
	/* readchar */
	case 43:
		param[0].xword = read_char();
		break;
	/* readchar(f) */
	case 44:
		loadfile((word) READING, &t1, &t2, &fp);
		if (t1 != TEXTF)
			errsignal(RTEILLIO);
		if ((ch = getc(fp)) == EOF)
			errsignal(RTEIOERR);
		param[0].xword = ch;
		break;
	/* readint */
	case 45:
		read_str(s);
		param[0].xword = atoi(s);
		break;
	
	case 46: /* readint(f) */
		loadfile((word) READING, &t1, &t2, &fp);
		if (t1 != TEXTF)
			errsignal(RTEILLIO);
		param[0].xword = readint(fp);
		break;
	/* readreal */
	case 47:
		read_str(s);
		param[0].xreal = (real)atof(s);
		break;
	/* readreal(f) */
	case 48:
		loadfile((word) READING, &t1, &t2, &fp);
		if (t1 != TEXTF)
			errsignal(RTEILLIO);
		param[0].xreal = (real)readreal(fp);
		break;
	/* getchar(f) */
	case 49:
		loadfile((word) READING, &t1, &t2, &fp);
		if (t1 != CHARF)
			errsignal(RTEILLIO);
		if ((ch = getc(fp)) == EOF)
			errsignal(RTEIOERR);
		param[0].xword = ch;
		break;
	/* getint(f) */
	case 50:
		loadfile((word) READING, &t1, &t2, &fp);
		if (t1 != INTF)
			errsignal(RTEILLIO);
		n = fread((char *) &param[ 0 ].xword, sizeof(word), 1, fp);
		if (n != 1)
			errsignal(RTEIOERR);
		break;
	/* getreal(f) */
	case 51:
		loadfile((word) READING, &t1, &t2, &fp);
		if (t1 != REALF)
			errsignal(RTEILLIO);
		n = fread((char *) &param[0].xreal, sizeof(real), 1, fp);
		if (n != 1)
			errsignal(RTEIOERR);
		break;
	/* getobject(f) */
	case 52:
		absent = TRUE;
		break;
	/* putchar(f) */
	case 53:
		loadfile((word) WRITING, &t1, &t2, &fp);
		if (t1 != CHARF)
			errsignal(RTEILLIO);
		ch = (char) param[ 0 ].xword;
		if (putc(ch, fp) == EOF)
			errsignal(RTEIOERR);
		break;
	/* putint(f) */
	case 54:
		loadfile((word) WRITING, &t1, &t2, &fp);
		if (t1 != INTF)
			errsignal(RTEILLIO);
		n = fwrite((char *) &param[ 0 ].xword, sizeof(word), 1, fp);
		if (n != 1)
			errsignal(RTEIOERR);
		break;
	/* putreal(f) */
	case 55:
		loadfile((word) WRITING, &t1, &t2, &fp);
		if (t1 != REALF)
			errsignal(RTEILLIO);
		n = fwrite((char *) &param[ 0 ].xreal, sizeof(real), 1, fp);
		if (n != 1)
			errsignal(RTEIOERR);
		break;
	/* putobject(f) */
	case 56:
	/* putstring(f) */
	case 57:
		absent = TRUE;
		break;
	/* writeln(f) */
	case 58:
		loadfile((word) WRITING, &t1, &t2, &fp);
		if (t1 != TEXTF)
			errsignal(RTEILLIO);
		if (putc('\n', fp) == EOF)
			errsignal(RTEIOERR);
		if (fflush(fp))
			errsignal(RTEIOERR);
		break;
	/* writeln */
	case 59:
		write_str("\n");
		break;
	/* writechar(f) */
	case 60:
		loadfile((word) WRITING, &t1, &t2, &fp);
		if (t1 != TEXTF)
			errsignal(RTEILLIO);
		if (putc((char) param[0].xword, fp) == EOF) 
			errsignal(RTEIOERR);
		break;
	/* writechar */
	case 61:
		write_char((char) param[0].xword);
		break;
	/* writeint(f) */
	case 62:
		loadfile((word) WRITING, &t1, &t2, &fp);
		if (t1 != TEXTF)
			errsignal(RTEILLIO);
		writeint(param[0].xword, param[1].xword, fp);
		break;
	/* writeint */
	case 63:
		writeint(param[0].xword, param[1].xword, stdout);
		break;
	/* writereal0(f) */
	case 64:
	/* writereal1(f) */
	case 66:
	/* writereal2(f) */
	case 68:
		loadfile((word) WRITING, &t1, &t2, &fp);
		if (t1 != TEXTF)
			errsignal(RTEILLIO);
		writereal((double) param[0].xreal, param[1].xword,
							param[2].xword, fp);
		break;
	/* writereal0 */
	case 65:
	/* writereal1 */
	case 67:
	/* writereal2 */
	case 69:
		writereal((double) param[0].xreal, param[1].xword,
							param[2].xword, stdout);
		break;
	/* writestring(f) */
	case 70:
		loadfile((word) WRITING, &t1, &t2, &fp);
		if (t1 != TEXTF)
			errsignal(RTEILLIO);
		writestring(param[0].xword, param[1].xword, fp);
		break;
	/* writestring */
	case 71:
		writestring(param[0].xword, param[1].xword, stdout);
		break;
	/* open temporary file */
	case 72:
		genfileobj(TRUE, param[1].xword, tempfilename(),
							&param[0].xvirt, &t1);
		break;
	/* open external file */
	case 73:
		genfileobj(FALSE, param[1].xword, asciiz(&param[2].xvirt),
							&param[0].xvirt, &t1);
		break;
	/* eoln */
	case 74:
		param[0].xbool = lbool(testeoln(stdin));
		break;
	/* eoln(f) */
	case 75:
		loadfile((word) READING, &t1, &t2, &fp);
		if (t1 != TEXTF) errsignal(RTEILLIO);
		param[0].xbool = lbool(testeoln(fp));
		break;
	/* this coroutine */
	case 76:
		loadvirt(param[0].xvirt,
			thisp->prochead + M[thisp->prochead] + CHD);
		break;
	/* this process */
	case 77:
		{
			procaddr p;
			virtaddr v;
			p.node = ournode;
			p.pix  = thispix;
			p.mark = thisp->mark;
			mess2obj(thisp, &p, &v);
			param[0].xvirt = v;
		}
		break;
	/* reset:procedure(f:file) */
	case 78:
		if (member(&param[0].xvirt, &t2))
			reset(t2);
		else
			errsignal(RTEREFTN);
		break;
	/* rewrite:procedure(f:file) */
	case 79:
		if (member(&param[0].xvirt, &t2))
			rewrite(t2);
		else
			errsignal(RTEREFTN);
		break;
	/* unlink:procedure(f:file) */
	case 80:
		delete(&param[0].xvirt);
		break;
	/* seek:procedure(f:file, offset, base:integer) */
	case 81:
		storevirt(param[0].xvirt, currfile);
		loadfile((word) UPDATING, &t1, &t2, &fp);
		if (t1 != DIRECT)
			errsignal(RTEILLIO);
		if (fseek(fp, (long) param[1].xword, (int) param[2].xword))
			errsignal(RTEIOERR);
		break;
	/* getrec(f, a, n) */
	case 82:
		loadfile((word) UPDATING, &t1, &t2, &fp);
		if (t1 != DIRECT)
			errsignal(RTEILLIO);
		param[1].xword = directio(&param[0].xvirt, param[1].xword,
							(int (*)())fread, fp);
		break;
	/* putrec(f, a, n) */
	case 83:
		loadfile((word) UPDATING, &t1, &t2, &fp);
		if (t1 != DIRECT) errsignal(RTEILLIO);
		param[1].xword = directio(&param[0].xvirt, param[1].xword,
							(int (*)())fwrite, fp);
		break;
	/* position:function(f:file):real */
	case 84:
		storevirt(param[0].xvirt, currfile);
		loadfile((word) UPDATING, &t1, &t2, &fp);
		if (t1 != DIRECT)
			errsignal(RTEILLIO);
		param[1].xword =(int) ftell(fp);
		break;
	/* memavail:function:integer */
	case 98:
		param[0].xword = memavail();
		break;
	/* exec:function(c:arrayof char):integer */
	case 99:
		cp = asciiz(&param[0].xvirt);
		param[1].xword = system(cp);
		free(cp);
		break;
		
	default:
		nonstandard(nrproc);
		break;
	}
#if TRACE
	fflush(stdout);
#endif
	if (absent)
		errsignal(RTEUNSTP);
}

