#include <QtGui/QDialog>
#include <libconfig.h>

#include "vlp/config.h"
#include "OptionsDialog.h"

namespace loglan {
namespace vlp {
namespace dialog {

OptionsDialog::OptionsDialog(QString configFilePath, QWidget * parent)
	: QDialog(parent)
{
	setupUi(this);
	
	connect(buttonBox, SIGNAL(accepted()), this, SLOT(accept()));
	connect(buttonBox, SIGNAL(rejected()), this, SLOT(reject()));
	
	loadConfig(configFilePath.toStdString().c_str());
}

OptionsDialog::~OptionsDialog()
{
}

/**
 * Adds IP address to the configuration.
 * Additional window is displayed to add address to the list
 */
void OptionsDialog::on_addConnectionButton_clicked()
{
	QDialog d(this, Qt::Dialog);
	QLabel lab("IP Address:", &d);
	QLineEdit ed("", &d);
	QPushButton ob("", &d);
	QPushButton cb("", &d);

	if (connections) {
		ob.setGeometry(30, 60, 80, 30);
		ob.setText("Ok");
		ob.setDefault(TRUE);
		lab.setGeometry(10, 10, 60, 30);
		lab.setText("Address");
		ed.setGeometry(70, 10, 140, 30);
		cb.setGeometry(130, 60, 80, 30);
		cb.setText("Cancel");
		d.resize(240, 100);
		connect(&ob, SIGNAL(clicked()), &d, SLOT(accept()));
		connect(&cb, SIGNAL(clicked()), &d, SLOT(reject())); 
		if (d.exec())
			if (strcmp(ed.text().toAscii().data(), "") != 0) {
				connections->addItem(ed.text());
			}
	}
}

/**
 * Deletes current address from available connections.
 */
void OptionsDialog::on_delConnectionButton_clicked()
{
	if (connections->currentRow() != -1) {
		connections->takeItem(connections->currentRow());
	}
}

/**
 * Loads configuration from the given file.
 *
 * @see QKernel::LoadConfig
 *
 * @param fname Filename of the configuration file.
 */
void OptionsDialog::loadConfig(const char * fname)
{
	loglan::vlp::Config config;
	config.load(fname);

	nodeNumber->setValue(config.getNodeNumber());
	switch(config.getConnectionType()) {
	case loglan::vlp::EXPLICIT:
		explicitConnectionMode->setEnabled(true);
		registrationConnectionMode->setDisabled(true);
		break;
	case loglan::vlp::REGISTER:
		registrationConnectionMode->setEnabled(true);
		explicitConnectionMode->setDisabled(true);
		break;
	}

	for (auto host : config.getHosts()) {
		connections->addItem(host.c_str());
	}

	programsDirectory->setText(config.getProgramDir());
}

void OptionsDialog::saveConfig(QString fname)
{
	saveConfig(fname.toStdString().c_str());
}

void OptionsDialog::saveConfig(const char * fname)
{
	loglan::vlp::Config config;

	config.setProgramDir(programsDirectory->text().toStdString().c_str());
	config.setNodeNumber(nodeNumber->value());
	config.setConnectionType(
		explicitConnectionMode->isChecked() ? EXPLICIT : REGISTER
	);
	
	std::vector<std::string> hosts;
	for(int i = 0; i < connections->count(); i++) {
		hosts.push_back(connections->item(i)->text().toStdString());
	}
	config.setHosts(hosts);
	
	config.save(fname);
}

}
}
}
